#include "kz_jwt.h"

KZ_BEGIN_EXTERN_C

sr_xavp_t ** kz_get_keys_crt();
extern gen_lock_t * kz_keys_lock;
extern htable_api_t kz_ht_api;
extern char* last_payload_result;

KZ_END_EXTERN_C

static inline pid_t gettid(void)
{
  return syscall(SYS_gettid);
}

#include <string>
#include <iostream>
#include <sstream>
#include <iomanip>
#include <unistd.h>
#include <thread>
#include <chrono>
#include <memory>
#include <sys/wait.h>
#include <type_traits>
#include <cstdlib>
#include <cxxabi.h>
#include <regex>
#include <functional>
#include <atomic>
#include <mutex>
#include <random>
#include <regex>


typedef signed char		int8_t;

#include "include/kz_log.hpp"
#include "include/json.hpp"
#include "include/jwt.hpp"
#include "include/uuid.hpp"

using json = nlohmann::json;

typedef std::map<std::string, std::string> keys;

std::string rsa_pub_key = R"(-----BEGIN PUBLIC KEY-----
MIIBITANBgkqhkiG9w0BAQEFAAOCAQ4AMIIBCQKCAQEAvVfvZdeV0i7XUu6ayVUc
ufg7wPYBkpL+wxAPUFOQBjqkPFjl2rVGRBApfCliwnHDiyHqo8LRDUvmqhjIgz0c
kiabIxvcrAb4dG6FkFZeCKVtHjvMgSuBo67ButUdYcfdnV8LulI2JUNUm6xnTdbf
/1OuV9zH4fXxGVu9UdaldybrtlcrF2S23grURSs7IaeE5YtNIw7mwpwW23ICjQV+
94xfufTqaZtAp0Ybm2MMtCnGBgqi+FafiAj27lToCRZYTByxBmpyO6Z75Ntb8U2d
CZJi1ObLMvmA+XvcIrSo2MINoj5l6SFCCPv4oqGWYi2BW3lcEVXtVUsf02QgYQif
eQICCAE=
-----END PUBLIC KEY-----)";

struct kz_token {
	using json = nlohmann::json;
	using value_type = json;
	using object_type = json::object_t;
	using array_type = json::array_t;
	using string_type = std::string;
	using number_type = double;
	using integer_type = int64_t;
	using boolean_type = bool;

	static jwt::json::type get_type(const json &val) {
		using jwt::json::type;

		if (val.type() == json::value_t::boolean)
			return type::boolean;
		else if (val.type() == json::value_t::number_integer)
			return type::integer;
		else if (val.type() == json::value_t::number_float)
			return type::number;
		else if (val.type() == json::value_t::string)
			return type::string;
		else if (val.type() == json::value_t::array)
			return type::array;
		else if (val.type() == json::value_t::object)
			return type::object;
		else
			throw std::logic_error("invalid type");
	}

	static json::object_t as_object(const json &val) {
		if (val.type() != json::value_t::object)
			throw std::bad_cast();
		return val.get<json::object_t>();
	}

	static std::string as_string(const json &val) {
		if (val.type() != json::value_t::string)
			throw std::bad_cast();
		return val.get<std::string>();
	}

	static json::array_t as_array(const json &val) {
		if (val.type() != json::value_t::array)
			throw std::bad_cast();
		return val.get<json::array_t>();
	}

	static int64_t as_int(const json &val) {
		if (val.type() != json::value_t::number_integer)
			throw std::bad_cast();
		return val.get<int64_t>();
	}

	static bool as_bool(const json &val) {
		if (val.type() != json::value_t::boolean)
			throw std::bad_cast();
		return val.get<bool>();
	}

	static double as_number(const json &val) {
		if (val.type() != json::value_t::number_float)
			throw std::bad_cast();
		return val.get<double>();
	}

	static bool parse(json &val, std::string str) {
		val = json::parse(str.begin(), str.end());
		return true;
	}

	static std::string serialize(const json &val) {
		return val.dump();
	}
};

namespace kz
{
    namespace pkg
    {
		void reset_last_result()
		{
			if(last_payload_result) {
				pkg_free(last_payload_result);
			}
			last_payload_result = NULL;
		};

		void set_last_result(char *payload)
		{
			reset_last_result();
			int len = strlen(payload);
			char* value = (char*) pkg_malloc(len+1);
			if(value) {
				memcpy(value, payload, len);
				value[len] = '\0';
				last_payload_result = value;
			}
		}

		void set_last_json_result(json& jobj)
		{
			std::ostringstream jobjStream;
			jobjStream << jobj;
			std::string str_payload = jobjStream.str();

			set_last_result((char*) str_payload.c_str());
		}
    }

	namespace htable
	{
		std::string fetch(std::string table, std::string key)
		{
			std::string pem;
			str str_table = {(char*)table.c_str(), (int)table.length()};
			str str_key = {(char*)key.c_str(), (int)key.length()};
			ht_cell_t *cell = kz_ht_api.get_clone(&str_table, &str_key);
			if(cell != NULL) {
				pem= std::string(cell->value.s.s, cell->value.s.len);
				pkg_free(cell);
			}

			return pem;
		}
	}

	namespace xavp
	{
	// copied from core/xavp.c xavp_get_internal
	sr_xavp_t *get(str *name, sr_xavp_t **list)
	{
		sr_xavp_t *avp;
		unsigned int id;

		if(name==NULL || name->s==NULL)
			return NULL;
		id = get_hash1_raw(name->s, name->len);

		if(!list || !(*list))
			return NULL;
		avp = *list;
		while(avp)
		{
			if(avp->id==id && avp->name.len==name->len
					&& strncmp(avp->name.s, name->s, name->len)==0)
			{
				return avp;
			}
			avp = avp->next;
		}
		return NULL;
	}

	sr_xavp_t *store(str *name, sr_xval_t *val, sr_xavp_t **list)
	{
		sr_xavp_t *cur;

		if(val==NULL)
			return NULL;

		/* Find the current value */
		cur = get(name, list);
		if(cur==NULL) {
			return xavp_add_value(name, val, list);
		} else {
			return xavp_set_value(name, 0, val, list);
		}
	}

	}

	namespace pk
	{
		void fetch(const char *KeyId)
		{
		}

		void store(const char *KeyId, const char *Pem)
		{
		}

		void create(const char *KeyId, const char *Pem)
		{
		}
	}

	namespace jwt
	{
		//using decoded_jwt = ::jwt::decoded_jwt<kz_token>;
		using rsa =  ::jwt::algorithm::rsa;
		using rs256 = ::jwt::algorithm::rs256;
		using json = nlohmann::json;
		using default_clock  = ::jwt::default_clock;
		using date = std::chrono::system_clock::time_point::clock;
		using verifier = ::jwt::verifier<default_clock, kz_token>;
		using builder = ::jwt::builder<kz_token>;
		using claim = ::jwt::basic_claim<kz_token>;

		bool is_jwt(std::string& token)
		{
			try
			{
				auto decoded = ::jwt::decode<kz_token>(token);

				for (auto &e : decoded.get_header_claims())
					AMQP_DBG << e.first << " = " << e.second.to_json();
				for (auto &e : decoded.get_payload_claims())
					AMQP_DBG << e.first << " = " << e.second.to_json();

			}
			catch(const std::invalid_argument& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
				return false;
			}
			catch(const std::runtime_error& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
				return false;
			}
			catch(const std::exception& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
				return false;
			}

			return true;
		}

		bool is_valid(std::string& token)
		{

			try
			{
				auto decoded = ::jwt::decode<kz_token>(token);
				if (!decoded.has_key_id()) {
					AMQP_DBG << "NO KID";
					return false;
				}
				if (!decoded.has_algorithm()) {
					AMQP_DBG << "NO ALG";
					return false;
				}
				if (!decoded.has_issuer()) {
					AMQP_DBG << "NO ISS";
					return false;
				}
				if (decoded.get_issuer().compare("kazoo") != 0) {
					AMQP_DBG << "=/ ISS => " << decoded.get_issuer();
					return false;
				}
				if (decoded.get_algorithm().compare("RS256") != 0) {
					AMQP_DBG << "=/ ALG => " << decoded.get_algorithm();
					return false;
				}

				std::string pem = htable::fetch("jwt.keys", decoded.get_key_id());
				if (pem.empty()) {
					AMQP_DBG << "KID " << decoded.get_key_id() << " not found";
					return false;
				}

				rsa key = rs256(pem);
				auto ver = verifier({})
								.with_issuer("kazoo")
								.allow_algorithm(key);

				ver.verify(decoded);

				for (auto &e : decoded.get_header_claims())
					AMQP_DBG << e.first << " = " << e.second.to_json();
				for (auto &e : decoded.get_payload_claims())
					AMQP_DBG << e.first << " = " << e.second.to_json();

			}
			catch(const std::invalid_argument& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
				return false;
			}
			catch(const std::runtime_error& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
				return false;
			}
			catch(const std::exception& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
				return false;
			}

			return true;
		}

		json header(std::string& token)
		{

			auto _header = json::object();
			try
			{
				auto decoded = ::jwt::decode<kz_token>(token);

				for (auto &e : decoded.get_header_claims()) {
					const auto k = e.first;
					const auto v = e.second;
					_header[k] = v.to_json();
				}

			}
			catch(const std::invalid_argument& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}
			catch(const std::runtime_error& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}
			catch(const std::exception& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}

			return _header;
		}

		json payload(std::string& token)
		{
			auto _payload = json::object();

			try
			{
				auto decoded = ::jwt::decode<kz_token>(token);

				for (auto &e : decoded.get_payload_claims())
					_payload[e.first] = e.second.to_json();

			}
			catch(const std::invalid_argument& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}
			catch(const std::runtime_error& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}
			catch(const std::exception& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}

			return _payload;
		}

		/*
		void set_payload(std::string& token)
		{
			auto j = payload(token);
			pkg::set_last_json_result(j);
		}
		*/

		json token(std::string& token)
		{
			auto tok = json::object();
			auto _header = json::object();
			auto _payload = json::object();
			try
			{
				auto decoded = ::jwt::decode<kz_token>(token);

				tok["header"] = _header;
				tok["payload"] = _payload;
				tok["signature"] = decoded.get_signature();

				for (auto &e : decoded.get_header_claims())
					_header[e.first] = e.second.to_json();

				for (auto &e : decoded.get_payload_claims())
					_payload[e.first] = e.second.to_json();

			}
			catch(const std::invalid_argument& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}
			catch(const std::runtime_error& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}
			catch(const std::exception& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}

			return tok;
		}

		std::string signature(std::string& token)
		{
			std::string sig;

			try
			{
				auto decoded = ::jwt::decode<kz_token>(token);
				sig = decoded.get_signature();

			}
			catch(const std::invalid_argument& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}
			catch(const std::runtime_error& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}
			catch(const std::exception& exc)
			{
				AMQP_DBG << "error validating token - " << exc.what() << " => " << token;
			}

			return sig;
		}

		std::string create(std::string payload)
		{
			using sec = std::chrono::seconds;
			using min = std::chrono::minutes;

			json jobj;

			std::string kid, priv_pem, pub_pem, issuer, aud, scope;
			std::string token;

			try
			{
				jobj = json::parse(payload.c_str());

				kid = htable::fetch("jwt.keys", "jwt.id");
				priv_pem = htable::fetch("jwt.keys", "jwt.private");
				pub_pem = htable::fetch("jwt.keys", "jwt.public");
				issuer = htable::fetch("jwt.keys", "jwt.issuer");
				aud = htable::fetch("jwt.keys", "jwt.aud");
				scope = htable::fetch("jwt.keys", "jwt.scope");

				if (kid.empty()) {
					AMQP_DBG << "no key identifier";
					return token;
				}

				if (priv_pem.empty()) {
					AMQP_DBG << "no private key";
					return token;
				}

				if (pub_pem.empty()) {
					AMQP_DBG << "no public key";
					return token;
				}

				if (issuer.empty()) {
					AMQP_DBG << "default issuer";
					issuer = "kazoo";
				}

				/*
				if (aud.empty()) {
					AMQP_DBG << "default audience";
					aud = node();
				}

				if (scope.empty()) {
					AMQP_DBG << "default scope";
					scope = "sip:call";
				}
				*/

				rsa key = rs256(pub_pem, priv_pem);

				const auto time = date::now();
				builder build;

				for(auto& item : jobj.items())
					build.set_payload_claim(item.key(), item.value());

				build.set_type("JWT")
					 .set_key_id(kid)
					 .set_issuer(issuer)
					 .set_expires_at(time + sec{15} + min{2});

				if (!aud.empty()) build.set_audience(aud);
				if (!scope.empty()) build.set_payload_claim("scope", claim(scope));

				token = build.sign(key);
			}
			catch(const std::exception& exc)
			{
				AMQP_DBG << "error parsing body - " << exc.what() << " => " << payload;
				return token;
			}

			return token;

		}

		void sign(const char *KeyId, const char *Pem)
		{
		}

	}

}

KZ_EXTERN_C int kz_jwt_validate(sip_msg_t* msg, char* token_p)
{
	str token_s = STR_NULL;
	if (fixup_get_svalue(msg, (gparam_p)token_p, &token_s) != 0) {
		LM_ERR("cannot get payload value\n");
		return -1;
	}

	std::string token(token_s.s, token_s.len);
	return kz::jwt::is_valid(token) ? 1 : -1;

}

KZ_EXTERN_C int kz_is_jwt(sip_msg_t* msg, char* token_p)
{
	str token_s = STR_NULL;
	if (fixup_get_svalue(msg, (gparam_p)token_p, &token_s) != 0) {
		LM_ERR("cannot get payload value\n");
		return -1;
	}

	std::string token(token_s.s, token_s.len);
	return kz::jwt::is_jwt(token) ? 1 : -1;

}

KZ_EXTERN_C int kz_jwt_payload(sip_msg_t* msg, char* token_p)
{
	str token_s = STR_NULL;
	if (fixup_get_svalue(msg, (gparam_p)token_p, &token_s) != 0) {
		LM_ERR("cannot get payload value\n");
		return -1;
	}

	std::string token(token_s.s, token_s.len);

	if (!kz::jwt::is_jwt(token)) return -1;

	json jobj = kz::jwt::payload(token);
	kz::pkg::set_last_json_result(jobj);

	return 1;

}

KZ_EXTERN_C int kz_jwt_token(sip_msg_t* msg, char* token_p)
{
	str token_s = STR_NULL;
	if (fixup_get_svalue(msg, (gparam_p)token_p, &token_s) != 0) {
		LM_ERR("cannot get payload value\n");
		return -1;
	}

	std::string token(token_s.s, token_s.len);

	if (!kz::jwt::is_jwt(token)) return -1;

	json jobj = kz::jwt::token(token);
	kz::pkg::set_last_json_result(jobj);

	return 1;

}

KZ_EXTERN_C int kz_jwt_header(sip_msg_t* msg, char* token_p)
{
	str token_s = STR_NULL;
	if (fixup_get_svalue(msg, (gparam_p)token_p, &token_s) != 0) {
		LM_ERR("cannot get payload value\n");
		return -1;
	}

	std::string token(token_s.s, token_s.len);

	if (!kz::jwt::is_jwt(token)) return -1;

	json jobj = kz::jwt::header(token);
	kz::pkg::set_last_json_result(jobj);

	return 1;

}
